@@ -18,6 +18,7 @@ except ImportError:
     HAVE_THREADS = False
 
 TIMEOUT = 1
+MS_WINDOWS = (os.name == 'nt')
 
 Py_REF_DEBUG = hasattr(sys, 'gettotalrefcount')
 
@@ -120,12 +121,12 @@ else:
         except (ValueError, resource_error):
             pass
 
-def spawn_python(*args):
+def spawn_python(*args, **kwargs):
     args = (sys.executable,) + args
     return subprocess.Popen(args,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
-                            stderr=subprocess.STDOUT)
+                            stderr=subprocess.STDOUT, **kwargs)
 
 def expected_traceback(lineno1, lineno2, header, min_count=1):
     regex = header
@@ -138,17 +139,17 @@ def expected_traceback(lineno1, lineno2, header, min_count=1):
 
 @contextmanager
 def temporary_filename():
-   filename = tempfile.mktemp()
-   try:
-       yield filename
-   finally:
-       try:
-           os.unlink(filename)
-       except OSError:
-           pass
+    filename = tempfile.mktemp()
+    try:
+        yield filename
+    finally:
+        try:
+            os.unlink(filename)
+        except OSError:
+            pass
 
 class FaultHandlerTests(unittest.TestCase):
-    def get_output(self, code, filename=None):
+    def get_output(self, code, filename=None, **kwargs):
         """
         Run the specified code in Python (in a new child process) and read the
         output from the standard error or from a file (if filename is set).
@@ -160,7 +161,7 @@ class FaultHandlerTests(unittest.TestCase):
         """
         code = dedent(code).strip()
         with SuppressCrashReport():
-            process = spawn_python('-c', code)
+            process = spawn_python('-c', code, **kwargs)
         stdout, stderr = process.communicate()
         exitcode = process.wait()
         output = re.sub(br"\[\d+ refs\]\r?\n?", b"", stdout).strip()
@@ -175,8 +176,9 @@ class FaultHandlerTests(unittest.TestCase):
                         output)
         return output.splitlines(), exitcode
 
-    def check_fatal_error(self, code, line_number, name_regex,
-                          filename=None, all_threads=True, other_regex=None):
+    def check_error(self, code, line_number, fatal_error,
+                          filename=None, all_threads=True, other_regex=None,
+                          thread_name=r"python([23](\.\d)?)?", **kwargs):
         """
         Check that the fault handler for fatal errors is enabled and check the
         traceback from the child process output.
@@ -184,37 +186,61 @@ class FaultHandlerTests(unittest.TestCase):
         Raise an error if the output doesn't match the expected format.
         """
         if all_threads:
-            header = 'Current thread XXX (most recent call first)'
+            if sys.version_info[:2] == (2, 6):
+                thread_name = 'python'
+            if sys.platform.startswith('linux') and thread_name is not None:
+                header = r'Current thread XXX <{0}> \(most recent call first\)'.format(thread_name)
+            else:
+                header = r'Current thread XXX \(most recent call first\)'
         else:
-            header = 'Stack (most recent call first)'
+            header = 'Stack \(most recent call first\)'
         regex = """
-            ^Fatal Python error: {name}
+            ^{fatal_error}
 
             {header}:
               File "<string>", line {lineno} in <module>
             """
         regex = dedent(regex).format(
             lineno=line_number,
-            name=name_regex,
-            header=re.escape(header)).strip()
+            fatal_error=fatal_error,
+            header=header).strip()
         if other_regex:
             regex += '|' + other_regex
-        output, exitcode = self.get_output(code, filename)
+        output, exitcode = self.get_output(code, filename, **kwargs)
         output = '\n'.join(output)
         self.assertRegex(output, regex)
         self.assertNotEqual(exitcode, 0)
 
+    def check_fatal_error(self, code, line_number, name_regex, **kw):
+        fatal_error = 'Fatal Python error: %s' % name_regex
+        self.check_error(code, line_number, fatal_error, **kw)
+
+    def check_windows_exception(self, code, line_number, name_regex, **kw):
+        fatal_error = 'Windows (fatal )?exception: %s' % name_regex
+        self.check_error(code, line_number, fatal_error, **kw)
+
     @skipIf(sys.platform.startswith('aix'),
             "the first page of memory is a mapped read-only on AIX")
     def test_read_null(self):
-        self.check_fatal_error("""
-            import faulthandler
-            faulthandler.enable()
-            faulthandler._read_null()
-            """,
-            3,
-            # Issue #12700: Read NULL raises SIGILL on Mac OS X Lion
-            '(?:Segmentation fault|Bus error|Illegal instruction)')
+        if not MS_WINDOWS:
+            self.check_fatal_error("""
+                import faulthandler
+                faulthandler.enable()
+                faulthandler._read_null()
+                """,
+                3,
+                # Issue #12700: Read NULL raises SIGILL on Mac OS X Lion
+                '(?:Segmentation fault'
+                    '|Bus error'
+                    '|Illegal instruction)')
+        else:
+            self.check_windows_exception("""
+                import faulthandler
+                faulthandler.enable()
+                faulthandler._read_null()
+                """,
+                3,
+                'access violation')
 
     def test_sigsegv(self):
         self.check_fatal_error("""
@@ -297,14 +323,23 @@ class FaultHandlerTests(unittest.TestCase):
     @skipIf(not hasattr(faulthandler, '_stack_overflow'),
             'need faulthandler._stack_overflow()')
     def test_stack_overflow(self):
-        self.check_fatal_error("""
-            import faulthandler
-            faulthandler.enable()
-            faulthandler._stack_overflow()
-            """,
-            3,
-            '(?:Segmentation fault|Bus error)',
-            other_regex='unable to raise a stack overflow')
+        if not MS_WINDOWS:
+            self.check_fatal_error("""
+                import faulthandler
+                faulthandler.enable()
+                faulthandler._stack_overflow()
+                """,
+                3,
+                '(?:Segmentation fault|Bus error)',
+                other_regex='unable to raise a stack overflow')
+        else:
+            self.check_windows_exception("""
+                import faulthandler
+                faulthandler.enable()
+                faulthandler._stack_overflow()
+                """,
+                3,
+                'stack overflow')
 
     def test_gil_released(self):
         self.check_fatal_error("""
@@ -337,6 +372,23 @@ class FaultHandlerTests(unittest.TestCase):
             'Segmentation fault',
             all_threads=False)
 
+    def test_enable_env_var(self):
+        output, exitcode = self.get_output("""
+            import faulthandler
+            print(faulthandler.is_enabled())
+            """,
+            env=dict(os.environ, PYTHONFAULTHANDLER='x'))
+        self.assertEqual(output, ["True"])
+        self.assertEqual(exitcode, 0)
+
+        output, exitcode = self.get_output("""
+            import faulthandler
+            print(faulthandler.is_enabled())
+            """,
+            env=dict(os.environ, PYTHONFAULTHANDLER=''))
+        self.assertEqual(output, ["False"])
+        self.assertEqual(exitcode, 0)
+
     def test_disable(self):
         code = """
             import faulthandler
@@ -454,6 +506,41 @@ class FaultHandlerTests(unittest.TestCase):
         self.assertEqual(trace, expected)
         self.assertEqual(exitcode, 0)
 
+    @skipIf(sys.platform != 'linux2', 'thread name printing is only supported on Linux')
+    def test_thread_name_when_set(self):
+        self.check_fatal_error("""
+            import faulthandler
+            faulthandler.enable()
+            import ctypes
+            libc = ctypes.cdll.LoadLibrary("libc.so.6")
+            buf = ctypes.create_string_buffer("some_name")
+            # PR_SET_NAME
+            libc.prctl(15, buf)
+            faulthandler._sigsegv()
+            """,
+            8,
+            'Segmentation fault',
+            thread_name="some_name"
+        )
+
+    @skipIf(sys.platform != 'linux2', 'thread name printing is only supported on Linux')
+    def test_thread_name_when_empty(self):
+        " Will not print the <thread_name> if no name is set "
+        self.check_fatal_error("""
+            import faulthandler
+            faulthandler.enable()
+            import ctypes
+            libc = ctypes.cdll.LoadLibrary("libc.so.6")
+            buf = ctypes.create_string_buffer("")
+            # PR_SET_NAME
+            libc.prctl(15, buf)
+            faulthandler._sigsegv()
+            """,
+            8,
+            'Segmentation fault',
+            thread_name=None
+        )
+
     @skipIf(not HAVE_THREADS, 'need threads')
     def check_dump_traceback_threads(self, filename):
         """
@@ -501,13 +588,13 @@ class FaultHandlerTests(unittest.TestCase):
         else:
             lineno = 11
         regex = """
-            ^Thread 0x[0-9a-f]+ \(most recent call first\):
+            ^Thread 0x[0-9a-f]+ (\<[\w\.]{{1,16}}\>\s)?\(most recent call first\):
             (?:  File ".*threading.py", line [0-9]+ in [_a-z]+
             ){{1,3}}  File "<string>", line 24 in run
               File ".*threading.py", line [0-9]+ in _?_bootstrap_inner
               File ".*threading.py", line [0-9]+ in _?_bootstrap
 
-            Current thread XXX \(most recent call first\):
+            Current thread XXX (\<[\w\.]{{1,16}}\>\s)?\(most recent call first\):
               File "<string>", line {lineno} in dump
               File "<string>", line 29 in <module>$
             """
@@ -573,7 +660,7 @@ class FaultHandlerTests(unittest.TestCase):
             count = loops
             if repeat:
                 count *= 2
-            header = r'Timeout \(%s\)!\nCurrent thread XXX \(most recent call first\):\n' % timeout_str
+            header = r'Timeout \(%s\)!\nCurrent thread XXX (\<[\w\.]{1,16}\>\s)?\(most recent call first\):\n' % timeout_str
             regex = expected_traceback(12, 23, header, min_count=count)
             self.assertRegex(trace, regex)
         else:
@@ -675,7 +762,7 @@ class FaultHandlerTests(unittest.TestCase):
         trace = '\n'.join(trace)
         if not unregister:
             if all_threads:
-                regex = 'Current thread XXX \(most recent call first\):\n'
+                regex = 'Current thread XXX (\<[\w\.]{1,16}\>\s)?\(most recent call first\):\n'
             else:
                 regex = 'Stack \(most recent call first\):\n'
             regex = expected_traceback(7, 28, regex)
